from abc import abstractmethod
from typing import Mapping, Sequence

from dagster._core.errors import DagsterCodeLocationLoadError
from dagster._core.host_representation.code_location import (
    CodeLocation,
)
from dagster._core.workspace.workspace import (
    CodeLocationEntry,
    CodeLocationStatusEntry,
    IWorkspace,
    location_status_from_location_entry,
)


class BaseDaemonWorkspace(IWorkspace):
    """An IWorkspace that lazily loads the list of locations in the workspace. Each daemon in
    the `dagster-daemon` process has its own DaemonWorkspace for thread safety, with a shared
    (thread-safe) GrpcServerRegistry so that if the process spins up its own gRPC servers, they
    the server processes can be shared across multiple daemons.

    Both the list of locations and the CodeLocation objects are cached until the daemon
    code calls cleanup() on the DaemonWorkspace - daemons are responsible for doing this
    periodically whenever they might want to check for code updates and workspace.yaml updates.
    """

    def __init__(self):
        self._location_entries = None

    def __enter__(self):
        return self

    def get_workspace_snapshot(self) -> Mapping[str, CodeLocationEntry]:
        if self._location_entries is None:
            self._location_entries = self._load_workspace()
        return dict(self._location_entries)

    def get_code_location_statuses(self) -> Sequence[CodeLocationStatusEntry]:
        if self._location_entries is None:
            self._location_entries = self._load_workspace()
        return [
            location_status_from_location_entry(entry) for entry in self._location_entries.values()
        ]

    @abstractmethod
    def _load_workspace(self) -> Mapping[str, CodeLocationEntry]:
        pass

    def get_workspace_copy_for_iteration(self):
        return DaemonIterationWorkspace(self.get_workspace_snapshot())

    def get_code_location(self, location_name: str) -> CodeLocation:
        if self._location_entries is None:
            self._location_entries = self._load_workspace()

        if location_name not in self._location_entries:
            raise DagsterCodeLocationLoadError(
                f"Location {location_name} does not exist in workspace",
                load_error_infos=[],
            )

        location_entry = self._location_entries[location_name]

        if location_entry.load_error:
            raise DagsterCodeLocationLoadError(
                f"Failure loading {location_name}: {location_entry.load_error}",
                load_error_infos=[location_entry.load_error],
            )

        if not location_entry.code_location:
            raise DagsterCodeLocationLoadError(
                f"Location {location_name} is still loading",
                load_error_infos=[],
            )

        return location_entry.code_location

    def cleanup(self, cleanup_locations: bool) -> None:
        if self._location_entries is not None:
            if cleanup_locations:
                for location_entry in self._location_entries.values():
                    if location_entry.code_location:
                        location_entry.code_location.cleanup()
            self._location_entries = None

    def __exit__(self, exception_type, exception_value, traceback):
        self.cleanup(cleanup_locations=True)


class DaemonIterationWorkspace(BaseDaemonWorkspace):
    """A copy of the main workspace's locations that can be called from a background thread
    in a daemon without worrying that the main thread will clean up the locations underneath us.
    Analagous to WorkspaceRequestContext in the webserver.

    Daemons that call this should be careful to set cleanup_locations=False when calling cleanup
    on the parent workspace that get_workspace_copy_for_iteration() was called on to create
    this workspace.
    """

    def __init__(self, location_entries_copy):
        self._location_entries_copy = location_entries_copy
        super().__init__()

    def _load_workspace(self) -> Mapping[str, CodeLocationEntry]:
        return self._location_entries_copy
